import os

class WindowsPath(object):
    """
    Processes Windows paths for easy conversion into cwRsync compatible
    format.  This involves replacing drive specification into
    /cygdrive/[drive_lettter] format, and replacing backward slashes with
    forward slashes.  While backward slashes will allow correctly locating the
    filesystem resource locally, it most likely will result in incorrect
    destinations on the remote side, as backward slashes will be interpreted as
    part of the name itself.
    """
    def __init__(self, path):
        """Path is any (absolute or relative) Windows path."""
        if os.path.isabs(path):
            drive, tail = os.path.splitdrive(path)
            self.drive = drive.rstrip(':')
            self.tail = tail.lstrip('\\').lstrip('/')
        else:
            self.drive = ''
            self.tail = path.lstrip('\\').lstrip('/')

    def to_cygpath(self):
        """
        This is where the rsync safe conversion takes place.  If there is a
        space in the path, encapsulates the path with quotes to assure the path
        will not be misinterpreted as being two or more paths.
        """
        drive = os.path.join('\\cygdrive', self.drive) if self.drive else ''
        path = os.path.join(drive, self.tail)
        path = path.replace('\\', '/')
        if " " in path:
            path = '"%s"' % path
        return path

def regularize_path(path):
    if os.name == 'nt':
        return WindowsPath(path).to_cygpath()
    else:
        return path



import os
from pkg_resources import get_distribution, DistributionNotFound

from .multyvac import (
    Multyvac,
    MultyvacError,
    RequestError
)

# Automatically determine the version of the package
try:
    _dist = get_distribution('multyvac')
    if not __file__.startswith(os.path.join(_dist.location, 'multyvac')):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'Only available when installed with setup.py'
else:
    # FIXME: If there are two egg-info folders (an older version failed to
    # uninstall), then the older one is chosen, which is incorrect.
    __version__ = _dist.version

# Default Multyvac singleton
_multyvac = Multyvac()

on_multyvac = _multyvac.on_multyvac
send_log_to_support = _multyvac.send_log_to_support

# Job methods that have been elevated to top level
from .job import JobError
modulemgr = _multyvac.job._modulemgr
get = _multyvac.job.get
get_by_name = _multyvac.job.get_by_name
list = _multyvac.job.list
kill = _multyvac.job.kill
shell_submit = _multyvac.job.shell_submit
submit = _multyvac.job.submit

# All other modules
from .config import ConfigError
config = _multyvac.config
from .volume import SyncError
volume = _multyvac.volume
layer = _multyvac.layer
cluster = _multyvac.cluster
api_key = _multyvac.api_key



import datetime
import json
import logging
import os
import random
import subprocess
import sys
import time

try:
    from cloghandler import (
        ConcurrentRotatingFileHandler as RotatingFileHandler
    )
except ImportError:
    from logging.handlers import RotatingFileHandler 

import requests
from requests.exceptions import ConnectionError
from util.cygwin import regularize_path

class MultyvacError(Exception):
    pass

class RequestError(MultyvacError):
    """Exception class for errors when making web requests to the
    Multyvac API."""

    def __init__(self, http_status_code, code, message, hint=None,
                 retry=False):
        Exception.__init__(self, http_status_code, code, message, hint, retry)
        self.http_status_code = http_status_code
        self.code = code
        self.message = message
        self.hint = hint
        self.retry = retry
    
    def __str__(self):
        return '%s (Code: %s Hint: %s)' % (self.message, self.code, self.hint)
    
    def __repr__(self):
        return 'RequestError({code}, "{message}", {hint})'.format(
                    code=self.code,
                    message=self.message,
                    hint=self.hint,
                )

class SyncError(MultyvacError):
    """Encapsulates errors when making rsync requests to Multyvac."""
    
    def __init__(self, exit_status, message):
        Exception.__init__(self, exit_status, message)
        self.exit_status = exit_status
        self.message = message
        
    def __repr__(self):
        return 'SyncError({exit_status}, "{message}")'.format(
                    exit_status=self.exit_status,
                    message=self.message,
                )

class Multyvac(object):
    """
    Multyvac

    The primary object for interacting with the Multyvac API.
    All Multyvac modules are exposed through this.
    """
    
    _ASK_GET = 'GET'
    _ASK_POST = 'POST'
    _ASK_PUT = 'PUT'
    
    def __init__(self, api_key=None, api_secret_key=None, api_url=None):
        self._session = requests.session()
        
        from .config import ConfigModule
        # Note: At this time, the rest of the Multyvac modules have not been
        # initialized. So the constructor should not do anything that requires
        # any other modules (ie. Do not use the ApiKey module).
        self.config = ConfigModule(self, api_key, api_secret_key, api_url)
        
        if os.name == 'nt':
            self._rsync_bin = os.path.join(self.config.get_multyvac_path(),
                                           'bin/rsync.exe')
            self._ssh_bin = os.path.join(self.config.get_multyvac_path(),
                                         'bin/ssh.exe')
        else:
            self._rsync_bin = 'rsync'
            self._ssh_bin = 'ssh'
        
        # Must be after config
        self._setup_logger()
        
        from .job import JobModule
        self.job = JobModule(self)
        from .layer import LayerModule
        self.layer = LayerModule(self)
        from .volume import VolumeModule
        self.volume = VolumeModule(self)
        from .cluster import ClusterModule
        self.cluster = ClusterModule(self)
        from .api_key import ApiKeyModule
        self.api_key = ApiKeyModule(self)
    
    def _setup_logger(self):
        """
        Sets up a rotating file logger.
        TODO: Have config option for printing to screen.
        """
        
        logs_path = os.path.join(self.config.get_multyvac_path(), 'log')
        if not os.path.exists(logs_path):
            self.config._create_path_ignore_existing(logs_path)
        log_path = os.path.join(logs_path, 'multyvac.log')
            
        self._logger = logging.getLogger('multyvac')
        self._logger.setLevel(logging.INFO)
    
        if os.name == 'nt':
            from logging import FileHandler
            try:
                handler = FileHandler(log_path, 'a')
            except Exception as e:
                print >> sys.stderr, 'Could not open logging file handler:', e
        else:
            try:
                handler = RotatingFileHandler(log_path, 'a', 1024*1024, 10)
            except OSError as e:
                if e.errno == 13:
                    # Assume the permission denied is because the log is owned by
                    # a different user (probably due to sudo). Use a different log
                    # file with the user's uid embedded.
                    # TODO: send_log_to_support() does not handle this case.
                    log_path = os.path.join(logs_path,
                                            'multyvac.%s.log' % os.getuid())
                    handler = RotatingFileHandler(log_path, 'a', 1024*1024, 10)
        # TODO: Fixing the permissions here only helps so much. If the user ran
        # this as sudo, the rotated logs will be owned by sudo's target user.
        # Our only option is to modify RotatingFileHandler to try to set the
        # calling user as the owner.
        self.config._fix_permission(log_path)
        lock_path = os.path.join(logs_path, 'multyvac.lock')
        self.config._fix_permission(lock_path)
        
        formatter = logging.Formatter(
            '[%(asctime)s] - [%(levelname)s] - %(name)s: %(message)s'
        )
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)
    
    def _get_session_method(self, method):
        """
        Returns a function that can be used to make an API request.
        :param method: The HTTP verb to be used by the request.
        """
        if method == self._ASK_POST:
            return self._session.post
        elif method == self._ASK_GET:
            return self._session.get
        elif method == self._ASK_PUT:
            return self._session.put
        else:
            raise KeyError('Unknown method "%s"' % method)
    
    def _log_ask(self, method, uri, params, data, headers, files):
        """Use this to log a request.  It only logs params and data elements
        that are not overly large to prevent filling up the log."""
        self._logger.info('%s request to %s with params %r data %r files %r',
                          method,
                          uri,
                          self._log_ask_element(params),
                          self._log_ask_element(data),
                          [path for path, _ in files.values()] if files else None)
    
    def _log_ask_element(self, ele):
        """Recurses into dict and list objects replacing elements that are too
        large for a log file. This way we still see small elements, but filter
        our large things like stdin."""
        max_element_byte_size = 150
        ele_size = sys.getsizeof(ele)
        if isinstance(ele, dict):
            d = {}
            for k, v in ele.items():
                d[k] = self._log_ask_element(v)
            return d
        elif isinstance(ele, (tuple, list)):
            return [self._log_ask_element(v) for v in ele]
        elif ele_size > max_element_byte_size:
            return 'Too large to log: %s bytes' % ele_size
        else:
            return ele
    
    def _ask(self, method, uri, auth=None, params=None, data=None,
             headers=None, files=None, content_type_json=False):
        """
        Makes an HTTP request to Multyvac.
        
        :param method: HTTP Verb.
        :param uri: Resource path.
        :param auth: Authentication override. If not specified, falls back to
            any credentials available in the MultyvacConfigModule.
        :param params: Query string parameters specified as a dict.
        :param data: Specify as dict. If not a JSON request, the dict is
            form encoded and put into the body (typical POST). If its a JSON
            request, then the data is serialized JSON in the request body.
        :param files: List of tuples [(path, content), ...] specifying files
            that should be uploaded as part of a multipart request.
        :param content_type_json: Whether the request body should be encoded as
            JSON, along with the appropriate content-type header. If False,
            regular form encoding is used.
        """
        if content_type_json:
            headers = headers or {}
            headers['content-type'] = 'application/json'
            final_data = json.dumps(data)
        else:
            final_data = data
        
        attempt = 0
        max_attempts = 5
        while True:
            self._log_ask(method, uri, params, data, headers, files)
            try:
                r = self._ask_helper(method,
                                     uri,
                                     auth=auth,
                                     params=params,
                                     data=final_data,
                                     headers=headers,
                                     files=files)
                return r
            except (RequestError, ConnectionError) as e:
                attempt += 1
                min_delay = 1.0
                if isinstance(e, RequestError) and e.http_status_code == 429:
                    # Add another attempt if error was due to rate limiting.
                    # This also increases the range of exponential backoff.
                    max_attempts += 1
                    min_delay = 5.0
                if ((isinstance(e, ConnectionError) or e.retry)
                        and attempt < max_attempts):
                    delay = max(2**attempt * random.random(), min_delay)
                    self._logger.info('Request failed. Retrying in %.1fs',
                                      delay)
                    time.sleep(delay)
                    continue
                else:
                    raise
    
    def _ask_helper(self, method, uri, auth, params, data, headers, files):
        """See _ask()."""
        
        if not auth:
            auth = self.config.get_auth()
        
        r = self._get_session_method(method)(
                self.config.api_url + uri,
                auth=auth,
                params=params,
                data=data,
                headers=headers,
                files=files,
            )
        
        try:
            obj = r.json()
        except ValueError:
            if r.status_code >= 500:
                # Retry on 5** error codes returning non-JSON (probably HTML)
                # Bad gateway is a common example where we want to do this.
                raise RequestError(r.status_code,
                                   None,
                                   r.text,
                                   retry=True)
            else:
                # Unexpected error
                raise RequestError(r.status_code,
                                   None,
                                   'Could not parse body',
                                   hint=r.text)
        if 'error' in obj:
            raise RequestError(r.status_code,
                               obj['error']['code'],
                               obj['error']['message'],
                               obj['error'].get('hint'),
                               obj['error'].get('retry'))
        
        return obj
    
    def _sync_up(self, local_path, remote_address, remote_path, port):
        """Sync from local path to Multyvac."""
        dest = 'multyvac@{address}:{path}'.format(address=remote_address,
                                                  path=remote_path)
        return self._sync(local_path, dest, port)
    
    def _sync_down(self, remote_address, remote_path, port, local_path):
        """Sync from Multyvac to local path."""
        src = 'multyvac@{address}:{path}'.format(address=remote_address,
                                                 path=remote_path)
        return self._sync(src, local_path, port)
    
    def _sync(self, src, dest, port):
        """Sync from source to destination using rsync."""
        
        attempt = 0
        max_attempts = 5
        while True:
            try:
                return self._sync_helper(src, dest, port)
            except SyncError as e:
                attempt += 1
                # connection refused errors return 255
                if e.exit_status == 255 and attempt < max_attempts:
                    delay = 2**attempt * random.random()
                    self._logger.info('Sync failed. Retrying in %.1fs',
                                      delay)
                    time.sleep(delay)
                    continue
                else:
                    raise
    
    def _sync_helper(self, src, dest, port):
        """The port might apply to either the src or the dest, depending on
        which one is remote."""
        
        on_windows = os.name == 'nt'
        
        cmd = ('{rsync_bin} -avz -L -e "{ssh_bin} -o UserKnownHostsFile=/dev/null '
               '-o StrictHostKeyChecking=no -p {port} -i {key_path}" {chmod} '
               '{src} {dest}'.format(
                    rsync_bin=self._rsync_bin,
                    ssh_bin=self._ssh_bin,
                    port=port,
                    key_path=regularize_path(self.config.path_to_private_key()),
                    chmod='--chmod=u+rwx' if on_windows else '',
                    src=src,
                    dest=dest)
               )
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            # close_fds not supported on Windows with stdout/stderr redirection
            close_fds=not on_windows,
            shell=True,
        )
        _, stderr = p.communicate()
        if p.poll() != 0:
            self._logger.info('Sync had error:\n%s',
                              stderr)
            raise SyncError(p.poll(), stderr)        
    
    def on_multyvac(self):
        """Returns True if this process is currently running on Multyvac."""
        return os.getenv('ON_MULTYVAC') == 'true'
    
    def send_log_to_support(self):
        """Sends this machine's log file to Multyvac support."""
        log_path = os.path.join(self.config.get_multyvac_path(),
                                'log/multyvac.log')
        if os.path.exists(log_path):
            with open(log_path) as f:
                files = {'file': ('multyvac.log', f.read())}
            self._ask(self._ASK_POST,
                      '/report/client_log/',
                      files=files)
            return True
        else:
            return False

class MultyvacModule(object):
    """All modules should extend this class."""
    def __init__(self, multyvac):
        self.multyvac = multyvac
        logger_name = self.__class__.__name__.lower()[:-len('module')]
        self._logger = logging.getLogger('multyvac.%s' % logger_name)

    @staticmethod
    def clear_null_entries(d):
        for k, v in d.items():
            if v is None:
                del d[k]

    @staticmethod
    def convert_str_to_datetime(s):
        try:
            return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            # FIXME
            return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')

    @staticmethod
    def check_success(r):
        return r['status'] == 'ok'
    
    @staticmethod
    def is_iterable_list(obj):
        return hasattr(obj, '__iter__')


class MultyvacModel(object):
    def __init__(self, multyvac=None, **kwargs):
        if multyvac:
            self.multyvac = multyvac
        else:
            raise Exception('Needs multyvac object for now')
            
    def __str__(self):
        return repr(self)




import getpass
import json
import os
import platform
import sys
import urllib
import zipfile

def report_install(m):
    data = {'hostname': platform.node(),
            'platform': platform.platform(),
            'processor': platform.processor(),
            'language': 'python',
            'language_extras': json.dumps({
                'version': platform.python_version(),
                'implementation': platform.python_implementation(),
                'architecture': platform.machine(),
                'pyexe_build': platform.architecture()[0],
                }),
            }
    m._ask(m._ASK_POST,
           '/report/install/',
           data=data)

def is_windows():
    return os.name == 'nt'

def install_multyvac_win32_bin(m):
    path = os.path.join(m.config.get_multyvac_path(), 'bin')
    if not (os.path.exists(path) and os.listdir(path)):
        print 'Downloading and installing dependencies, ssh and rsync... '
        m.config._create_path_ignore_existing(path)
        # TODO: Probably need to add retry logic
        tmp_file_path, _ = urllib.urlretrieve(
            m.config.api_url + '/setup/multyvac-win32-bin.zip'
        )
        z = zipfile.ZipFile(tmp_file_path)
        z.extractall(path)
    else:
        print 'Dependencies, ssh and rsync, are already installed.'

def report_install(m):
    import json
    import platform
    data = {'hostname': platform.node(),
            'platform': platform.platform(),
            'processor': platform.processor(),
            'language': 'python',
            'language_extras': json.dumps({
                'version': platform.python_version(),
                'implementation': platform.python_implementation(),
                'architecture': platform.machine(),
                'pyexe_build': platform.architecture()[0],
                }),
            }
    m._ask(m._ASK_POST,
           '/report/install/',
           data=data)

def main():
    
    print 'Welcome to Multyvac.\n'
    print ('If you input your Multyvac username and password, an api key will '
           'be automatically fetched for this machine.') 
    
    username = raw_input('Username: ')
    password = getpass.getpass('Password: ')
    
    from .multyvac import Multyvac, RequestError
    m = Multyvac()
    try:
        api_keys = m.api_key.list(username=username, password=password)
    except RequestError as e:
        print >> sys.stderr, ('Bad username and password. '
                              'Please run setup again.')
        sys.exit(1)
    except Exception as e:
        print >> sys.stderr, 'Unknown error: ', e
        sys.exit(1)
    
    for api_key in api_keys:
        if api_key.active:
            print 'Your machine will use api key %s' % api_key.id
            m.config.set_key(api_key.id,
                             api_key.secret_key)
            m.config.save_to_disk()
            if is_windows():
                install_multyvac_win32_bin(m)
            print 'Success'
            try: report_install(m)
            except: pass
            sys.exit(0)
    else:
        print >> sys.stderr, 'Your account has no active api keys.'
        sys.exit(1)

    

if __name__ == '__main__':
    main()



import base64
import copy
try: import cPickle as pickle
except: import pickle
try: from cStringIO import StringIO
except: import StringIO
from functools import partial
import inspect
import socket
import subprocess
import time

from .multyvac import (
    Multyvac,
    MultyvacError,
    MultyvacModel,
    MultyvacModule,
    RequestError,
)
from .util import preinstalls
from .util.cygwin import regularize_path
from .util.module_dependency import ModuleDependencyAnalyzer

class JobError(MultyvacError):
    """Exception class for errors encountered by a job."""
    pass

class Job(MultyvacModel):
    """Represents a Multyvac Job and its associated operations."""
    
    status_waiting = 'waiting'
    status_queued = 'queued'
    status_processing = 'processing'
    
    status_done = 'done'
    status_error = 'error'
    status_killed = 'killed'
    status_stalled = 'stalled'
    
    finished_statuses = [status_done,
                         status_error,
                         status_killed,
                         status_stalled]
    
    def __init__(self, jid, **kwargs):
        MultyvacModel.__init__(self, **kwargs)
        self.jid = jid
        
        self.cmd = kwargs.get('cmd')
        self.core = kwargs.get('core')
        self.created_at = kwargs.get('created_at')
        self.multicore = kwargs.get('multicore')
        self.name = kwargs.get('tags', {}).get('name')
        self.status = kwargs.get('status')
        self.tags = kwargs.get('tags')
    
        result = kwargs.get('result')
        self.result_type = kwargs.get('result_type')
        if result and self.result_type  == 'pickle':
            self.result = pickle.loads(base64.b64decode(result))
        elif result and self.result_type == 'binary':
            self.result = base64.b64decode(result)
        else:
            self.result = result
        self.return_code = kwargs.get('return_code')
        
        self.started_at = kwargs.get('started_at')
        self.finished_at = kwargs.get('finished_at')
        self.runtime = kwargs.get('runtime')
        self.queue_delay = kwargs.get('queue_delay')
        self.overhead_delay = kwargs.get('overhead_delay')
        
        self.cputime_user = kwargs.get('collected', {}).get('cputime_user')
        self.cputime_system = kwargs.get('collected', {}).get('cputime_system')
        self.memory_failcnt = kwargs.get('collected', {}).get('memory_failcnt')
        self.memory_max_usage = kwargs.get('collected', {}).get('memory_max_usage')
        self.ports = kwargs.get('collected', {}).get('ports')
        self.stderr = kwargs.get('stderr')
        self.stdout = kwargs.get('stdout')
    
    def get_result(self, raise_on_error=True):
        """
        Better than using the result attribute directly.
        
        First, it waits for the job to finish processing. Once finished,
        returns the result. If the job encountered an error (or stalled/killed)
        a JobError exception is raised.
        
        :param raise_on_error: If set to False, an exception is returned,
            rather than raised, in the event of an error.
        """
        self.wait()
        if self.status != self.status_done:
            e = JobError(self.stderr)
            if raise_on_error:
                raise e
            else:
                return e
        else:
            return self.result
    
    def kill(self):
        """
        :param jid: Can be a number or list of numbers.
        """
        
        data = {'jid': self.jid}
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job/kill',
                               data=data)
        return MultyvacModule.check_success(r)
    
    def wait(self, status=finished_statuses, timeout=None):
        """
        Wait for the job to reach the specified status.
        
        :param status: Can be a status string or a list of strings.
        
        Returns the status if the job reaches it, otherwise on a timeout
        returns False.
        """
        start_time = time.time()
        poll_period = 1
        max_poll_period = 10
        while True:
            current_time = time.time()
            time_elapsed = current_time - start_time
            if timeout and time_elapsed > timeout:
                return False
            elif self.status == status or self.status in status:
                return self.status
            else:
                sleep_time = (min(poll_period, timeout - time_elapsed)
                              if timeout else poll_period)
                time.sleep(sleep_time)
                if poll_period < max_poll_period:
                    poll_period += 0.5
                self.update()
    
    def wait_for_open_port(self, port, timeout=None):
        """
        Blocks until it detects that port has been opened by the job.
        Returns a dict of the external address and port combination (will
        be different than the input port due to NAT).
        
        If port is 22, also returns path to identity file, and username.
        
        :param timeout: The amount of time to wait for the job to start.
        
        Once started, the job must open the port within 10 seconds. Otherwise,
        False is returned.
        """
        port = str(port)
        self.wait(self.finished_statuses + [self.status_processing], timeout)
        if self.status == self.status_processing:
            attempt = 1
            max_attempts = 10
            while attempt < max_attempts and self.status == Job.status_processing:
                if not self.ports or not self.ports.get('tcp', {}).get(port):
                    # Wait for SSH to start
                    time.sleep(2.0)
                    attempt += 1
                    self.update()
                else:
                    d = copy.copy(self.ports.get('tcp', {}).get(port))
                    if port == '22':
                        d['username'] = 'multyvac'
                        d['identity_file'] = self.multyvac.config.path_to_private_key()
                    return d
            else:
                return False
        else:
            return False
    
    def open_ssh_console(self):
        """Opens an SSH console to a running job. If a job is queued, waits
        indefinitely until it is processing."""
        
        if self.wait_for_open_port(22):
            info = self.ports.get('tcp', {}).get('22')
            address = info['address']
            port = info['port']
            cmd = ('{ssh_bin} -o UserKnownHostsFile=/dev/null '
                   '-o StrictHostKeyChecking=no -X -p {port} -i {key_path} '
                   ' multyvac@{address}'.format(
                        ssh_bin=self.multyvac._ssh_bin,
                        port=port,
                        key_path=regularize_path(self.multyvac.config.path_to_private_key()),
                        address=address,)
                   )
            p = subprocess.Popen(cmd, shell=True)
            p.wait()
        
    def run_command(self, cmd):
        """
        Runs the specified command over ssh. Blocks until the job has begun
        processing.
        
        If successful, returns (stdout, stderr). Otherwise, returns False,
        which happens most notably if the job has already finished by the
        time this command is run.
        """
        
        if self.wait_for_open_port(22):
            info = self.ports.get('tcp', {}).get('22')
            address = info['address']
            port = info['port']
            cmd = ('{ssh_bin} -o UserKnownHostsFile=/dev/null '
                   '-o StrictHostKeyChecking=no -p {port} -i {key_path} '
                   ' multyvac@{address} {cmd}'.format(
                        ssh_bin=self.multyvac._ssh_bin,
                        port=port,
                        key_path=regularize_path(self.multyvac.config.path_to_private_key()),
                        address=address,
                        cmd=cmd)
                   )
            p = subprocess.Popen(cmd,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 shell=True)
            return p.communicate()
        else:
            self.multyvac.job._logger.info('Cannot SSH into finished job')
            return False
    
    def update(self):
        """Updates this Job object with the latest version available of itself
        from Multyvac."""
        j = self.multyvac.job.get(self.jid)
        self.__dict__ = j.__dict__
        return self.status

    def __repr__(self):
        if self.name:
            return '%s(%s, name=%s)' % (self.__class__.__name__,
                                        repr(self.jid), repr(self.name))
        else:
            return '%s(%s)' % (self.__class__.__name__, repr(self.jid))


class JobModule(MultyvacModule):
    """Most of JobModule's methods are exposed directly through ``multyvac``.
    For example, ``multyvac.submit()``."""

    def __init__(self, *args, **kwargs):
        MultyvacModule.__init__(self, *args, **kwargs)
        self._modulemgr = ModuleDependencyAnalyzer()
        preinstalled_modules = [name for name, _ in preinstalls.modules]
        self._modulemgr.ignore(preinstalled_modules)
    
    def _normalize_vol(self, vol):
        if isinstance(vol, basestring):
            return [{'name': vol}]
        elif isinstance(vol, dict):
            return [vol]
        elif hasattr(vol, '__iter__'):
            if isinstance(vol[0], dict):
                return vol
            elif isinstance(vol[0], basestring):
                return [{'name': v} for v in vol]
            else:
                raise TypeError('vol must be a iterable of strings or dicts')
        else:
            raise TypeError('vol must be a string, dict, or iterable')
        
    def _normalize_layer(self, layer):
        if isinstance(layer, basestring):
            return [{'name':layer}]
        elif isinstance(layer, dict):
            return [layer]
        else:
            raise TypeError('layer must be a string or dict')

    def shell_submit(self, cmd, _name=None, _core='c1', _multicore=1,
                     _layer=None,  _vol=None, _env=None,
                     _result_source='stdout', _result_type='binary',
                     _max_runtime=None, _profile=False, _restartable=True,
                     _tags=None, _depends_on=None, _stdin=None):
        """
        Submit a job to Multyvac.
        
        :param cmd: The shell command to execute.
        :param _name: The string name to give the job.
        :param _core: The core to use.
        :param _multicore: The number of cores to assign to the job.
        :param _layer: The name of the layer to use.
        :param _vol: The name or a list of names of volumes to mount.
        :param _env: A dictionary of any environment variables that should be
            set in the shell running the job.
        :param _result_source: The source of the result. By default it's the
            standard output of the job. But it can also be set to 'file:{path}'
            where {path} should be replaced by the location of a file that will
            contain the job's result.
        :param _result_type: The format of the job's result. By default, it is
            set to binary, which means Multyvac cannot interpret its contents.
        :param _max_runtime: The maximum number of seconds this job should be
            allowed to run before it is forcibly killed.
        :param _profile: Not implemented yet.
        :param _restartable: If a server running the job fails unexpectedly,
            can this job be safely restarted?
        :param _tags: A dict mapping keys to values of arbitrary data. Good for
            storing job metadata.
        :param _depends_on: Not implemented.
        :param _stdin: The standard input that should be piped into the job.
        
        :returns: Job id.
        """
        
        job = {
               'cmd': cmd,
               'name': _name,
               'core': _core,
               'multicore': _multicore,
               'profile': _profile,
               'restartable': _restartable,
               'depends_on': _depends_on,
               'tags': _tags,
               'layer': self._normalize_layer(_layer) if _layer else None,
               'vol': self._normalize_vol(_vol) if _vol else None,
               'env': _env,
               'result_source': _result_source,
               'result_type': _result_type,
               'max_runtime': _max_runtime,
               }
        
        if _stdin:
            job['stdin'] = base64.b64encode(_stdin)
        
        MultyvacModule.clear_null_entries(job)
        
        payload = {'jobs': [job]}
        
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job',
                               data=payload,
                               content_type_json=True)
        return r['jids'][0]

    def _get_auto_module_volume_name(self):
        return 'auto-deps-%s' % socket.gethostname()

    def submit(self, f, *args, **kwargs):
        """
        Submit a Python function as a job to Multyvac.
        
        :param args: The positional arguments to pass into f.
        :param kwargs: The keywords arguments to pass info f. kwargs may also
            include special keys that are prefixed with '_'. See the arguments
            to :meth:`shell_submit` for more information.
            
        :returns: Job id.
        """
        
        f_kwargs = {}
        for k, v in kwargs.items():
            if not k.startswith('_'):
                f_kwargs[k] = v
                del kwargs[k]
        
        from .util.cloudpickle import CloudPickler
        
        s = StringIO()
        cp = CloudPickler(s, 2)
        cp.dump((f, args, f_kwargs))
        
        # Add modules
        for module in cp.modules:
            self._modulemgr.add(module.__name__)
        
        mod_paths = self._modulemgr.get_and_clear_paths()
        
        vol_name = self._get_auto_module_volume_name()
        if self._modulemgr.has_module_dependencies:
            v = self.multyvac.volume.get(vol_name)
            if not v:
                try:
                    self.multyvac.volume.create(vol_name, '/pymodules')
                except RequestError as e:
                    if 'name already exists' not in e.message:
                        raise
                v = self.multyvac.volume.get(vol_name)
            if mod_paths:
                v.sync_up(mod_paths, '')
            
        kwargs['_stdin'] = s.getvalue()
        kwargs['_result_source'] = 'file:/tmp/.result'
        kwargs['_result_type'] = 'pickle'
        if self._modulemgr.has_module_dependencies:
            kwargs.setdefault('_vol', []).append(vol_name)
        # Add to the PYTHONPATH if user is using it as well
        env = kwargs.setdefault('_env', {})
        if env.get('PYTHONPATH'):
            env['PYTHONPATH'] = env['PYTHONPATH'] + ':/pymodules'
        else:
            env['PYTHONPATH'] = '/pymodules'
            
        tags = kwargs.setdefault('_tags', {})
        # Make sure function name fits within length limit for tags
        fname = JobModule._func_name(f)
        if len(fname) > 100:
            fname = fname[:97] + '...'
        tags['fname'] = fname
        
        return self.shell_submit(
            'python -m multyvacinit.pybootstrap',
            **kwargs
        )

    def list(self,
             jid=None,
             name=None,
             limit=50,
             before=None,
             after=None,
             status=None):
        """
        Query jobs.
        
        :param jid: Can be a job id or list of job ids.
        :param name: Can be a name or list of names.
        :param limit: Maximum number of jobs to return.
        :param before: Only jobs that were created before this jid are returned.
        :param after: Only jobs that were created after this jid are returned.
        :param status: Only jobs with this status are returned.
            Can be a list of statuses.
        
        Use `before` and `after` to paginate through jobs efficiently. 
            
        :returns: A list of matching jobs.
        """
        params = {'jid': jid,
                  'name': name,
                  'limit': limit,
                  'before': before,
                  'after': after,
                  'status': status,
                  #'field': self.fields,
                  }
        MultyvacModule.clear_null_entries(params)
        
        return self._get(params)

    def get(self, jid):
        """
        Returns a Job with corresponding id.
        
        :param jid: Can be a job id or list of job ids.
        
        :returns: A list if input was a list, otherwise returns an object.
            Also, can return None if a single jid is requested, and it cannot
            be found.
        """
        iter_in = MultyvacModule.is_iterable_list(jid)
        params = {'jid': jid}
        if not iter_in:
            params['limit'] = 1
        jobs = self._get(params)
        if iter_in:
            return jobs
        else:
            return jobs[0] if jobs else None
    
    def get_by_name(self, name):
        """
        Get a job by name.
        
        :param name: The name of the job.
        
        :returns: The most recent job with the matching name. Use :meth:`list`
            to get a list of all matching jobs.
        """
        
        iter_in = MultyvacModule.is_iterable_list(name)
        if iter_in:
            raise ValueError('name can only be a string, not a list')
        
        params = {'name': name,
                  'limit': 1}
        jobs = self._get(params)
        
        return jobs[0] if jobs else None
    
    def _get(self, more_params):
        
        params = {#'field': self.fields
                  }
        params.update(more_params)
        
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/job',
                               params=params)
        for job in r['jobs']:
            job['created_at'] = MultyvacModule.convert_str_to_datetime(job['created_at'])
        return [Job(multyvac=self.multyvac, **job) for job in r['jobs']]

    @staticmethod
    def _func_name(func):
        """Return name of a callable (function, class, partial, etc.)"""
        module = ''
        if hasattr(func,'__module__'):
            module = (func.__module__ if func.__module__ else '__main__')
        # Return a human readable name associated with a function
        if inspect.ismethod(func):
            nme = '.'.join([module,func.im_class.__name__,func.__name__]) 
        elif inspect.isfunction(func):
            nme =  '.'.join([module,func.__name__])
        elif inspect.isbuiltin(func):
            return  '.'.join([module,func.__name__])
        elif isinstance(func,partial):
            return 'partial_of_' + JobModule._func_name(func.func)    
        elif inspect.isclass(func):
            nme = '.'.join([module,func.__name__])
            if hasattr(func, '__init__') and inspect.ismethod(func.__init__):            
                func = func.__init__
            else:
                return nme
        else:
            nme = 'type %s' % type(func)
            if hasattr(func, '__name__'):
                nme = '%s of %s' % (func.__name__, type(func))
            return nme
        nme +=  ' at ' + ':'.join([func.func_code.co_filename,
                                   str(func.func_code.co_firstlineno)])
        return nme

    def kill(self, jid):
        """
        Kills a job. If the job is queued, it will never run. If it's
        processing, it will be abruptly stopped. If it's already finished,
        nothing changes.
        
        :param jid: A job id or list of job ids to kill.
        """
        
        data = {'jid': jid}
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/job/kill',
                               data=data)
        return MultyvacModule.check_success(r)
